import json
import uuid
from collections.abc import AsyncGenerator, Sequence
from enum import Enum
from random import randrange
from typing import Any

import httpx
from fastapi import APIRouter, HTTPException, status
from fastapi.responses import StreamingResponse
from fastapi_pagination import Page
from fastapi_pagination.ext.sqlalchemy import paginate
from sqlmodel import col, select
from starlette.background import BackgroundTask

from app import crud
from app.api.deps import CurrentUser, SessionDep
from app.core.config import settings
from app.core.telemetry import telemetry_client
from app.fixtures import fake_sofia_chat_response
from app.models import (
    ChatCreate,
    ChatMessage,
    ChatQuery,
    ChatQueryPublic,
    ChatQueryUpdate,
    Collection,
    Document,
    MessageRequest,
    UserRole,
)
from app.utils import user_permissions

# This file is inspired from : https://gitlab.adullact.net/dgfip/projets-ia/caradoc/-/blob/dc346202070924f0fa64ebfd495429682f150722/api/app/routers/chat.py
router = APIRouter()


client = httpx.AsyncClient(timeout=settings.CHAT.request_timeout)


ChatEventType = Enum(
    "ChatEventType",
    [
        ("INFORMATION", "information"),
        ("REFORMULATED_QUERY", "reformulated_query"),
        ("DOCS", "docs"),
        ("CHUNK", "chunk"),
        ("CO2_EQ", "gCO2_eq"),
    ],
)


async def transform_sofia_chat_response(
    sofia_chat_response: AsyncGenerator[str], doc_to_map: Sequence[uuid.UUID]
) -> AsyncGenerator[str]:
    """
    Transform sofia_chat response in a format compatible with caradoc based chat
    """
    async for line in sofia_chat_response:
        # Response is in json
        res = {}
        try:
            res = json.loads(line)
        except Exception as exc:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Unable to parse response json from sofia_chat",
            ) from exc

        if ChatEventType.INFORMATION.value in res:
            yield f"""{
                json.dumps(
                    {
                        "event": ChatEventType.INFORMATION.value,
                        "data": {
                            "title": res[ChatEventType.INFORMATION.value]["title"],
                        },
                    }
                )
            }$$$\n"""
        elif ChatEventType.REFORMULATED_QUERY.value in res:
            yield f"""{
                json.dumps(
                    {
                        "event": "rephrased_query",
                        "data": {"query": res[ChatEventType.REFORMULATED_QUERY.value]},
                    }
                )
            }$$$\n"""
        elif ChatEventType.DOCS.value in res:
            document_list = res[ChatEventType.DOCS.value]
            viewer_sources = [
                {
                    "id": doc["document_id"] if len(doc_to_map) == 0 else str(doc_to_map[randrange(len(doc_to_map))]),
                    "content": doc["extrait"],
                    "score": doc["score"],
                    # Title and filename are not available for now, they'll be retrieve by the front from the DB
                    "filename": "",
                    "filetype": "pdf",
                    "page": doc["page"],
                }
                for index, doc in enumerate(document_list)
            ]
            yield f"""{
                json.dumps(
                    {
                        "event": "sources",
                        "data": {
                            "sources": viewer_sources,
                        },
                    }
                )
            }$$$\n"""
        elif ChatEventType.CHUNK.value in res:
            yield f"""{
                json.dumps(
                    {
                        "event": "content",
                        "data": {"content": res[ChatEventType.CHUNK.value]},
                    }
                )
            }$$$\n"""
        elif ChatEventType.CO2_EQ.value in res:
            yield f"""{
                json.dumps(
                    {
                        "event": "co2_eq",
                        "data": {"co2_eq": res[ChatEventType.CO2_EQ.value]},
                    }
                )
            }$$$\n"""


@router.post(
    "/query/{id}/message",
    response_description="Answer user prompt request",
)
async def process_message(
    session: SessionDep, current_user: CurrentUser, id: uuid.UUID, chat_message: ChatMessage
) -> Any:
    """Process the user prompt request and generate an accurate answer

    Args:
        message_request (MessageRequest): The data received from the client
        - collection_id (str, optional): The collection id
        - message (str): The message

    Returns:
        StreamingResponse: A stream of data containing the sources and the response message

    Raises:
        HTTPException

    """
    if not settings.CHAT.url and not settings.CHAT.fake:
        raise HTTPException(status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="Sofia chat not configured")

    with user_permissions(current_user):
        crud.read_generic_item(session=session, model=ChatQuery, item_id=id, check_edit_permission=True)
        # Retieve collection documents and map them to the response
        documents_to_map: Sequence[uuid.UUID] = []
        if settings.CHAT.fake:
            sql_req = select(Document.id).where(Document.collection_id == chat_message.collection_ids[0])
            documents_to_map = session.exec(sql_req).all()
        statement = select(Collection).where(col(Collection.id).in_(chat_message.collection_ids))
        collections = session.exec(statement).all()

        missing = set(chat_message.collection_ids).difference({x.id for x in collections})
        if any(missing):
            raise HTTPException(status_code=404, detail=f"Invalid collections: {missing}")

        cannot_read = [x.id for x in collections if not x.can_read]
        if any(cannot_read):
            raise HTTPException(status_code=400, detail=f"Not enough privileges to query {cannot_read}")

        if len({col.type for col in collections}) > 1:
            raise HTTPException(status_code=400, detail="Cannot query collections of different type")

        if settings.CHAT.fake:
            return StreamingResponse(
                transform_sofia_chat_response(fake_sofia_chat_response(), documents_to_map),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "X-Accel-Buffering": "no",
                },
            )
        elif settings.CHAT.url:
            message_request = MessageRequest(
                qdrant_collection_name=collections[0].qdrant_collection_name,
                query=chat_message.query,
                sources=[x.title for x in collections],
            )
            # TODO: verify arg type
            req = client.build_request(
                settings.CHAT.request_type,
                settings.CHAT.chat,
                data=message_request.model_dump_json(),  # type: ignore[arg-type]
            )
            # print(req)
            r = await client.send(req, stream=True)
            return StreamingResponse(
                transform_sofia_chat_response(r.aiter_lines(), documents_to_map),  # type: ignore
                background=BackgroundTask(r.aclose),
            )


@router.get(
    "/queries", response_description="Get all the chat queries (superadmin only)", response_model=Page[ChatQueryPublic]
)
async def get_chat_queries(session: SessionDep, current_user: CurrentUser) -> Any:
    if current_user.role != UserRole.superadmin:
        raise HTTPException(status_code=403, detail="Only superadmin can access this route")
    statement = select(ChatQuery).order_by(ChatQuery.updated_at.desc())  # type: ignore[attr-defined]
    return paginate(session, statement)


@router.post("/query", response_description="Create query, needed to generated a response", response_model=ChatQuery)
async def create_chat_query(session: SessionDep, current_user: CurrentUser, chat_query_in: ChatCreate) -> Any:
    with user_permissions(current_user):
        max_title_len = 50
        return crud.create_generic_item(
            session=session,
            model=ChatQuery,
            item=chat_query_in,
            user_id=current_user.id,
            title=chat_query_in.query
            if len(chat_query_in.query) < max_title_len
            else chat_query_in.query[:max_title_len] + "...",
        ).model_dump()


@router.get("/query/{id}", response_description="Get a speicfic query", response_model=ChatQuery)
async def get_chat_query(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
    with user_permissions(current_user):
        return crud.read_generic_item(session=session, model=ChatQuery, item_id=id)


@router.put("/query/{id}", response_model=ChatQuery)
async def update_chat_query(
    session: SessionDep, current_user: CurrentUser, id: uuid.UUID, chat_query_in: ChatQueryUpdate
) -> Any:
    with user_permissions(current_user):
        payload = crud.update_generic_item(
            session=session, model=ChatQuery, item_update=chat_query_in, item_id=id
        ).model_dump()
        telemetry_client.capture(current_user.id, event="chat-query", properties=payload)
        return payload


@router.delete("/query/{id}")
async def delete_chat_query(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
    with user_permissions(current_user):
        telemetry_client.capture(current_user.id, event="chat-delete", properties={"id": id})
        return crud.delete_generic_item(session=session, model=ChatQuery, item_id=id)
