import uuid
from io import BytesIO
from typing import Any, cast

import pikepdf
from fastapi import APIRouter, HTTPException, UploadFile, status
from fastapi_pagination import Page
from fastapi_pagination.ext.sqlalchemy import paginate
from sqlalchemy import and_, or_
from sqlmodel import select
from starlette.datastructures import Headers

from app import crud
from app.api.deps import CurrentUser, SessionDep
from app.core.config import settings
from app.core.qdrant import client as qdrant_client
from app.core.qdrant import delete_vectors
from app.core.telemetry import telemetry_client
from app.models import (
    ChatQuery,
    Collection,
    CollectionCreate,
    CollectionPublic,
    CollectionType,
    CollectionUpdate,
    Document,
    DocumentCreate,
    DocumentPublic,
    Message,
    UserRole,
)
from app.utils import user_permissions

router = APIRouter()


@router.get("/", response_model=Page[CollectionPublic])
def read_collections(session: SessionDep, current_user: CurrentUser) -> Any:
    """
    Retrieve collections.
    """
    with user_permissions(current_user):
        if current_user.role == UserRole.superadmin:
            statement = select(Collection)
        elif current_user.role == UserRole.groupadmin:
            statement = select(Collection).where(
                or_(Collection.type == CollectionType.public, Collection.group_id == current_user.group_id)  # type: ignore[arg-type]
            )
        else:
            statement = select(Collection).where(
                or_(
                    Collection.type == CollectionType.public,  # type: ignore[arg-type]
                    Collection.user_id == current_user.id,  # type: ignore[arg-type]
                    and_(Collection.type == CollectionType.restricted, Collection.group_id == current_user.group_id),  # type: ignore[arg-type]
                )
            )
        return paginate(session, statement.order_by(Collection.created_at.desc()))  # type: ignore[attr-defined]


@router.get("/{id}", response_model=CollectionPublic)
def read_collection(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
    """
    Get collection by ID.
    """
    with user_permissions(current_user):
        return crud.read_generic_item(session=session, model=Collection, item_id=id)  # .model_dump()


@router.post("/", response_model=CollectionPublic)
def create_collection(*, session: SessionDep, current_user: CurrentUser, collection_in: CollectionCreate) -> Any:
    """
    Create new collection.
    """
    with user_permissions(current_user):
        if (
            current_user.role not in (UserRole.superadmin, UserRole.groupadmin)
            and collection_in.type != CollectionType.personal
        ):
            raise HTTPException(status_code=403, detail="The user doesn't have enough privileges")

        # Check if collection with the same title does not exist already
        statement = select(Collection).where(Collection.title == collection_in.title)
        if session.exec(statement).first() is not None:
            raise HTTPException(status_code=400, detail="A collection with the given name already exists")

        telemetry_client.capture(current_user.id, event="collection-create", properties=collection_in.model_dump())
        return crud.create_generic_item(
            session=session,
            model=Collection,
            item=collection_in,
            user_id=current_user.id if collection_in.type == CollectionType.personal else None,
            group_id=current_user.group_id,
        ).model_dump()


@router.put("/{id}", response_model=CollectionPublic)
async def update_collection(
    *,
    session: SessionDep,
    current_user: CurrentUser,
    id: uuid.UUID,
    collection_in: CollectionUpdate,
) -> Any:
    """
    Update a collection.
    """
    with user_permissions(current_user):
        collection = cast(
            Collection,
            crud.read_generic_item(session=session, model=Collection, item_id=id, check_edit_permission=True),
        )
        if collection_in.title and collection_in.title is not collection.title:
            # The collection title is unique in the DB, it determines the s3 path and the behaviour of the chat
            # so it's not possible to change it currently
            raise HTTPException(status.HTTP_501_NOT_IMPLEMENTED, detail="Cannot modify collection title")
        # if qdrant_client and collection_in.title:
        #     if collection.title != collection_in.title:
        #         await update_collection_name(
        #             collection.qdrant_collection_name,
        #             collection_name=collection.title,
        #             new_collection_name=collection_in.title,
        #         )
        return crud.update_generic_item(
            session=session, model=Collection, item_update=collection_in, item_id=id
        ).model_dump()


@router.delete("/{id}")
async def delete_collection(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Message:
    """
    Delete a collection.
    """
    with user_permissions(current_user):
        collection = cast(
            Collection,
            crud.read_generic_item(session=session, model=Collection, item_id=id, check_edit_permission=True),
        )
        telemetry_client.capture(current_user.id, event="collection-delete", properties={"collection_id": id})
        # Delete all documents in s3
        if collection.path.exists():
            collection.path.rmtree()
        if qdrant_client:
            await delete_vectors(collection.qdrant_collection_name, collection_name=collection.title)
        return crud.delete_generic_item(session=session, model=Collection, item_id=id)


@router.get("/{id}/documents", response_model=Page[DocumentPublic])
def read_documents(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
    # Fetch collection to make sure the user can read it
    read_collection(session=session, current_user=current_user, id=id)
    with user_permissions(current_user):
        return paginate(session, select(Document).where(Document.collection_id == id))


# Specific router to handle file upload
def on_tus_upload_complete(
    session: SessionDep, current_user: CurrentUser, file_path: str, metadata: dict[Any, Any], file_size: int
) -> None:
    collection_id = metadata["collection_id"]
    with user_permissions(current_user):
        collection = cast(Collection, crud.read_generic_item(session=session, model=Collection, item_id=collection_id))
        if not (collection.can_read and (collection.can_edit or current_user.role == UserRole.contributor)):
            raise HTTPException(status_code=403, detail="The user doesn't have enough privileges")
        # We convert the file to a BinaryIO type needed by UploadFile type
        with open(file_path, "rb") as binary_file:
            headers = Headers({"content-type": metadata["filetype"]})
            fake_upload_file = UploadFile(
                file=binary_file, filename=metadata["filename"], size=file_size, headers=headers
            )
            doc = handle_pdf_upload(session, collection_id, fake_upload_file)
            telemetry_client.capture(current_user.id, event="document-create", properties=doc.model_dump())


def handle_pdf_upload(session: Any, collection_id: uuid.UUID, file: UploadFile) -> Document:
    """
    Save pdf data to the document table and pdf content to a S3 bucket.
    Linearize the document before if it's not already linearized.
    """
    # TODO: get n_pages
    try:
        doc = pikepdf.open(filename_or_stream=BytesIO(file.file.read()))
        if len(doc.pages) > settings.MAX_UPLOAD_PDF_PAGES:
            raise HTTPException(
                status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE,
                detail=f"Document must contains less than {settings.MAX_UPLOAD_PDF_PAGES} pages.",
            )
        should_linearize = not doc.is_linearized
        if doc.is_linearized and not doc.check_linearization():
            should_linearize = True
            print("Document content incorrect linearization data")
        # I was unable to work with metadata on PDF < 2.0
        # meta = doc.open_metadata()
        # doc_title = meta["dc:title"]
        doc_title = ""
        if "/Title" in doc.docinfo:
            doc_title = str(doc.docinfo["/Title"])
        # Linearize document on the fly
        if should_linearize:
            print(f"Document '{file.filename}' not linearized, convert it to a streamable format")
            stream = BytesIO()
            doc.save(filename_or_stream=stream, linearize=True)
            file.file = stream
            stream.seek(0)
            # @todo find a way to avoid the need to re read the stream in order to get the new size
            file.size = len(stream.read())
        # Important because minio needs it after
        file.file.seek(0)
    except Exception as exc:
        if not isinstance(exc, HTTPException):
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Erreur à l'ouverture du document pdf",
            ) from exc
        else:
            raise exc
    try:
        new_doc = DocumentCreate(
            collection_id=collection_id,
            filename=file.filename,
            content_length=file.size,
            n_pages=len(doc.pages),
            title=doc_title,
        )
        # Important : don't use create_generic_item because we want to commit
        # data only when file has been written to s3 to avoid potential race condition
        created_doc = Document.model_validate(new_doc)
        session.add(created_doc)
        # Use flush to insert the document in db without closing the transaction
        session.flush()
        session.refresh(created_doc)
        created_doc.path.write_bytes(file.file.read())
        assert created_doc.path.exists()
        session.commit()
        session.refresh(created_doc)
        return created_doc
    except Exception as exc:
        session.rollback()
        try:
            if created_doc.id:
                session.delete(created_doc)
                session.commit()
        except Exception as cleanup_error:
            print(f"Erreur lors du nettoyage : {cleanup_error}")
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Erreur lors de la sauvegarde du document pdf",
        ) from exc


@router.post("/{id}/document", response_model=list[DocumentPublic])
async def create_document(
    *, session: SessionDep, current_user: CurrentUser, id: uuid.UUID, files: list[UploadFile]
) -> Any:
    with user_permissions(current_user):
        collection = cast(Collection, crud.read_generic_item(session=session, model=Collection, item_id=id))
        if not (collection.can_read and (collection.can_edit or current_user.role == UserRole.contributor)):
            raise HTTPException(status_code=403, detail="The user doesn't have enough privileges")
        uploaded_documents = []
        for file in files:
            doc = handle_pdf_upload(session, id, file)
            telemetry_client.capture(current_user.id, event="document-create", properties=doc.model_dump())
            uploaded_documents.append(doc.model_dump())
        return uploaded_documents


@router.get("/{id}/chat_queries", response_model=list[ChatQuery])
async def read_queries(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
    with user_permissions(current_user):
        collection = cast(Collection, crud.read_generic_item(session=session, model=Collection, item_id=id))
        if not collection.can_read:
            raise HTTPException(status_code=403, detail="The user doesn't have enough privileges")
        return session.exec(
            select(ChatQuery).where(ChatQuery.collection_id == id).where(ChatQuery.user_id == current_user.id)
        ).all()
