import uuid
from random import randrange
from typing import Any, cast

from celery.result import AsyncResult
from fastapi import APIRouter, HTTPException, status
from fastapi.responses import FileResponse, Response
from fastapi_pagination import Page
from fastapi_pagination.ext.sqlalchemy import paginate
from sqlmodel import select

from app import crud
from app.api.deps import CurrentSuperAdmin, CurrentUser, CurrentUserWithQuery, SessionDep
from app.core.celery import celery_app
from app.core.config import settings
from app.core.qdrant import client as qdrant_client
from app.core.qdrant import delete_vectors, update_meta
from app.core.telemetry import telemetry_client
from app.models import (
    Document,
    DocumentPublic,
    DocumentUpdate,
    Message,
    ProcessingCreated,
    ProcessingStatus,
    ProcessingStatusPublic,
)
from app.utils import create_document_metadata, user_permissions

router = APIRouter()


# For superadmin to quickly check the last documents accessed
@router.get("/", response_model=Page[DocumentPublic])
def read_documents(session: SessionDep, current_user: CurrentSuperAdmin) -> Any:
    """
    Retrieve documents.
    """
    with user_permissions(current_user):
        return paginate(session, select(Document).order_by(Document.updated_at.desc()))  # type: ignore[attr-defined]


# Not used for now. Can be helpfull when importing document which already has a url
# @router.post("/", response_model=DocumentPublic)
# def create_document(*, session: SessionDep, current_user: CurrentUser, document_in: DocumentCreate) -> Any:
#     """
#     Create new document.
#     """
#     with user_permissions(current_user):
#         collection = cast(
#             Collection, crud.read_generic_item(session=session, model=Collection, item_id=document_in.collection_id)
#         )
#         if not (collection.can_read and (collection.can_edit or current_user.role == UserRole.contributor)):
#             raise HTTPException(status_code=403, detail="The user doesn't have enough privileges")
#         return crud.create_generic_item(session=session, model=Document, item=document_in).model_dump()


# Not used for now, it may be usefull though
# @router.get("/{id}", response_model=DocumentPublic)
# def read_document(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
#     """
#     Get document by ID.
#     """
#     with user_permissions(current_user):
#         return crud.read_generic_item(session=session, model=Document, item_id=id).model_dump()


# We need to craft our HEAD response in order to be compatible with PSPDFKit expectations for linearization
# And since the GET function is dynamic, FastAPI could not guess the content length
# https://github.com/fastapi/fastapi/issues/1773 helped a bit
@router.head("/{id}/file")
def get_head(session: SessionDep, current_user: CurrentUserWithQuery, id: uuid.UUID) -> Any:
    with user_permissions(current_user):
        doc = cast(Document, crud.read_generic_item(session=session, model=Document, item_id=id))
        if doc:
            res = Response(media_type="application/pdf")
            res.headers["Accept-ranges"] = "bytes"
            res.headers["Content-Length"] = str(doc.content_length)
            res.headers["Content-Range"] = "bytes 0-" + str(doc.content_length - 1) + "/" + str(doc.content_length)
            res.status_code = status.HTTP_206_PARTIAL_CONTENT
            return res
        else:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail="Aucun fichier avec correspond à l'identifiant",
            )


@router.get("/{id}/file")
async def get_file(session: SessionDep, current_user: CurrentUserWithQuery, id: uuid.UUID) -> Any:
    with user_permissions(current_user):
        doc = cast(Document, crud.read_generic_item(session=session, model=Document, item_id=id))
        if doc:
            return FileResponse(doc.path)
        else:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail="Aucun fichier ne correspond à cet identifiant",
            )


@router.put("/{id}", response_model=DocumentPublic)
async def update_document(
    *,
    session: SessionDep,
    current_user: CurrentUser,
    id: uuid.UUID,
    document_in: DocumentUpdate,
) -> Any:
    """
    Update a document.
    """
    with user_permissions(current_user):
        if document_in.meta and qdrant_client:
            doc = cast(
                Document,
                crud.read_generic_item(session=session, model=Document, item_id=id, check_edit_permission=True),
            )
            await update_meta(doc.qdrant_collection_name, document_in.meta, document_id=id)
        return crud.update_generic_item(
            session=session, model=Document, item_update=document_in, item_id=id
        ).model_dump()


@router.delete("/{id}")
async def delete_document(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Message:
    """
    Delete a document.
    """
    with user_permissions(current_user):
        doc = cast(
            Document, crud.read_generic_item(session=session, model=Document, item_id=id, check_edit_permission=True)
        )
        telemetry_client.capture(current_user.id, event="document-delete", properties={"id": id})
        # This will remove the pdf file and the metadata.yml if present
        doc.path.parent.rmtree()
        if qdrant_client:
            await delete_vectors(doc.qdrant_collection_name, document_id=id)
        return crud.delete_generic_item(session=session, model=Document, item_id=id)


@router.post("/{id}/process")
def process_document(
    session: SessionDep, current_user: CurrentUser, id: uuid.UUID, rerun: bool = False
) -> ProcessingCreated:
    """
    Process a document: send a celery task and register task id in db
    """
    with user_permissions(current_user):
        document: Document = cast(
            Document, crud.read_generic_item(session=session, model=Document, item_id=id, check_edit_permission=True)
        )
        if document.processing_task_id and not rerun:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Document already processed and rerun not requested"
            )
        if celery_app is None and not settings.CHAT.fake:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="Document processing unavailable"
            )

        # Create metadata.yaml
        create_document_metadata(document)

        # Send task
        task_id = uuid.uuid4()
        if celery_app is None or settings.CHAT.fake:
            # Register task id in db
            document.processing_task_id = task_id
            session.add(document)
            session.commit()
            return ProcessingCreated(task_id=task_id)

        kwargs = {
            "document_id": str(document.id),
            "collection_name": document.collection.title,
            "qdrant_collection_name": document.qdrant_collection_name,
        }
        celery_app.send_task("sofia.process_document", kwargs=kwargs, task_id=str(task_id))

        # Register task id in db
        document.processing_task_id = task_id
        session.add(document)
        session.commit()

        return ProcessingCreated(task_id=task_id)


@router.post("/{id}/process/status")
def update_processing_status(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> ProcessingStatusPublic:
    """
    Update status of document processing by asking REDIS or postgres db link to sofia document processing
    """
    with user_permissions(current_user):
        document: Document = cast(Document, crud.read_generic_item(session=session, model=Document, item_id=id))
        if celery_app is None and not settings.CHAT.fake:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="Document processing unavailable"
            )
        if not document.processing_task_id:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="No processing task id, you need to call /document/{id}/process before.",
            )
        # Return the value in DB if processing in ready state
        if document.processing_status in [ProcessingStatus.failure, ProcessingStatus.success]:
            return ProcessingStatusPublic(status=document.processing_status, details=document.processing_status_details)
        new_status = document.processing_status
        details = None
        if settings.CHAT.fake:
            possible_status = [
                ProcessingStatus.failure,
                ProcessingStatus.success,
                ProcessingStatus.pending,
                ProcessingStatus.started,
            ]
            new_status = ProcessingStatus(possible_status[randrange(len(possible_status))])
            details = "Status details"
        elif celery_app is None:  # only for mypy
            raise ValueError("Should never happen")
        else:
            result: AsyncResult[None] = AsyncResult(str(document.processing_task_id), backend=celery_app.backend)
            new_status = ProcessingStatus(result.status)
            details = str(result.info)

        if new_status is not document.processing_status:
            document.processing_status = new_status
            document.processing_status_details = str(details)  # Conversion needed by SQL to accept dict
            crud.update_generic_item(session=session, model=Document, item_update=document, item_id=id)

        return ProcessingStatusPublic(status=new_status, details=details)


# N.B.: Does not, maybe because the task is running in a solo pool
@router.post("/{id}/process/stop")
def stop_processing_document(session: SessionDep, current_user: CurrentUser, id: uuid.UUID) -> Any:
    with user_permissions(current_user):
        document: Document = cast(Document, crud.read_generic_item(session=session, model=Document, item_id=id))
        if not document.can_edit:
            raise HTTPException(status_code=403, detail="The user doesn't have enough privileges")
        if celery_app is None and not settings.CHAT.fake:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="Document processing unavailable"
            )
        if not document.processing_task_id:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="No processing task id, you need to call /document/{id}/process before.",
            )
        if celery_app is not None:  # for mypy
            celery_app.control.revoke(str(document.processing_task_id), terminate=True, signal="SIGKILL")
    return Message(message=f"Task {document.processing_task_id} revoked")
