import secrets
import sys
import warnings
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal
from urllib.parse import urljoin

from pydantic import (
    AnyUrl,
    BaseModel,
    BeforeValidator,
    Field,
    HttpUrl,
    PostgresDsn,
    StringConstraints,
    computed_field,
    model_validator,
)
from pydantic_core import MultiHostUrl
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self


def parse_cors(v: Any) -> list[str] | str:
    if isinstance(v, str) and not v.startswith("["):
        return [i.strip() for i in v.split(",")]
    elif isinstance(v, list | str):
        return v
    raise ValueError(v)


class CollectionType(str, Enum):
    personal = "personal"
    restricted = "restricted"
    public = "public"


# from https://github.com/pydantic/pydantic/discussions/7793
DomainName = Annotated[
    str, StringConstraints(pattern=r"^(?:[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$")
]


class SofiaChatSettings(BaseModel):
    url: AnyUrl | None = None
    chat_endpoint: str = "/chat/chat_sofia"
    search_endpoint: str = "/search/query"
    request_type: Literal["GET", "POST"] = "POST"
    request_timeout: int = 120
    fake: bool = False

    @property
    def chat(self) -> str:
        return urljoin(str(self.url), self.chat_endpoint)


class QdrantSettings(BaseModel):
    host: str | None = Field(default=None, alias="link")
    port: int = Field(default=6333)
    timeout: int = Field(default=30)
    collection_names: dict[CollectionType, str] = Field(
        default={
            CollectionType.public: "main",
            CollectionType.restricted: "restricted",
            CollectionType.personal: "personal",
        },
        alias="colnames",
    )
    collection_field: str = Field(
        default="metadata.collection",
        description="Field name in qdrant metadata that stores the document collection name",
        alias="colfield",
    )
    document_id_field: str = Field(
        default="metadata.document_id",
        description="Field name in qdrant metadata that stores the document collection name",
        alias="docfield",
    )


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        # Use top level .env file (one level above ./backend/)
        env_file="../.env",
        env_ignore_empty=True,
        extra="ignore",
        env_nested_delimiter="_",  # N.B.: fields inside the models with _ do not work since it is the delimiter
    )
    TESTING: bool = "pytest" in sys.modules
    API_V1_STR: str = "/api/v1"
    SECRET_KEY: str = secrets.token_urlsafe(32)
    # 60 minutes * 24 hours * 8 days = 8 days
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8
    FRONTEND_HOST: str = "http://localhost:5173"
    # TEMP : Causes weird CORS errors on sspcloud for some routes
    # TODO : need to investigate
    FRONTEND_DIST_PATH: Path | None = None  # Path("app/dist/")
    ENVIRONMENT: Literal["local", "staging", "production"] = "local"

    # S3 Config
    S3_IS_ENDPOINT_SECURE: bool = True
    S3_ENDPOINT: str
    S3_ACCESS_KEY: str
    S3_SECRET_KEY: str
    S3_SESSION_TOKEN: str | None = None
    S3_COLLECTIONS_BUCKET_NAME: str

    BACKEND_CORS_ORIGINS: Annotated[list[AnyUrl] | str, BeforeValidator(parse_cors)] = []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def all_cors_origins(self) -> list[str]:
        return [str(origin).rstrip("/") for origin in self.BACKEND_CORS_ORIGINS] + [self.FRONTEND_HOST]

    PROJECT_NAME: str = "Sofia-collections"
    SENTRY_DSN: HttpUrl | None = None
    POSTGRES_SERVER: str
    POSTGRES_PORT: int = 5432
    POSTGRES_USER: str
    POSTGRES_PASSWORD: str = ""
    POSTGRES_DB: str = ""

    @computed_field  # type: ignore[prop-decorator]
    @property
    def SQLALCHEMY_PG_URI(self) -> AnyUrl:
        return AnyUrl.build(
            scheme="postgresql+psycopg",
            username=self.POSTGRES_USER,
            password=self.POSTGRES_PASSWORD,
            host=self.POSTGRES_SERVER,
            port=self.POSTGRES_PORT,
        )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def SQLALCHEMY_DATABASE_URI(self) -> PostgresDsn:
        return MultiHostUrl.build(
            scheme="postgresql+psycopg",
            username=self.POSTGRES_USER,
            password=self.POSTGRES_PASSWORD,
            host=self.POSTGRES_SERVER,
            port=self.POSTGRES_PORT,
            path=self.POSTGRES_DB,
        )

    SMTP_TLS: bool = True
    SMTP_SSL: bool = False
    SMTP_PORT: int = 587
    SMTP_HOST: str | None = None
    SMTP_USER: str | None = None
    SMTP_PASSWORD: str | None = None
    # TODO: update type to EmailStr when sqlmodel supports it
    EMAILS_FROM_EMAIL: str | None = None
    EMAILS_FROM_NAME: str | None = None

    @model_validator(mode="after")
    def _set_default_emails_from(self) -> Self:
        if not self.EMAILS_FROM_NAME:
            self.EMAILS_FROM_NAME = self.PROJECT_NAME
        return self

    EMAIL_RESET_TOKEN_EXPIRE_HOURS: int = 48

    @computed_field  # type: ignore[prop-decorator]
    @property
    def emails_enabled(self) -> bool:
        return bool(self.SMTP_HOST and self.EMAILS_FROM_EMAIL)

    # TODO: update type to EmailStr when sqlmodel supports it
    EMAIL_TEST_USER: str = "test@example.com"
    # TODO: update type to EmailStr when sqlmodel supports it
    FIRST_SUPERUSER: str
    FIRST_SUPERUSER_PASSWORD: str
    DEFAULT_GROUP: str
    ACCEPTED_DOMAINS: list[DomainName] = []

    OPENID_CONFIG_URL: str | None = None
    KEYCLOAK_CLIENT_ID: str | None = None
    REDIS_HOST: str | None = None
    POSTHOG_KEY: str | None = Field(min_length=1, default=None)

    CHAT: SofiaChatSettings = SofiaChatSettings()

    QDRANT: QdrantSettings = QdrantSettings()

    MAX_UPLOAD_SIZE_IN_MB: int = 1024
    MAX_UPLOAD_PDF_PAGES: int = 1500

    def _check_default_secret(self, var_name: str, value: str | None) -> None:
        if value == "changethis":
            message = (
                f'The value of {var_name} is "changethis", for security, please change it, at least for deployments.'
            )
            if self.ENVIRONMENT == "local":
                warnings.warn(message, stacklevel=1)
            else:
                raise ValueError(message)

    @model_validator(mode="after")
    def _enforce_non_default_secrets(self) -> Self:
        self._check_default_secret("SECRET_KEY", self.SECRET_KEY)
        self._check_default_secret("POSTGRES_PASSWORD", self.POSTGRES_PASSWORD)
        self._check_default_secret("FIRST_SUPERUSER_PASSWORD", self.FIRST_SUPERUSER_PASSWORD)

        return self


settings = Settings()  # type: ignore
