import uuid
from typing import Any

from qdrant_client import AsyncQdrantClient, models

from app.core.config import settings

client = (
    AsyncQdrantClient(url=settings.QDRANT.host, port=settings.QDRANT.port, timeout=settings.QDRANT.timeout)
    if settings.QDRANT.host is not None
    else None
)
if client is None:
    print("Qdrant client not set")


def get_filter(collection_name: str | None, document_id: uuid.UUID | None) -> models.Filter:
    if collection_name is None and document_id is None:
        raise ValueError("document_id or collection_name must be provided")
    return models.Filter(
        must=(
            [
                models.FieldCondition(
                    key=settings.QDRANT.document_id_field,
                    match=models.MatchValue(value=str(document_id)),
                )
            ]
            if document_id is not None
            else [
                models.FieldCondition(
                    key=settings.QDRANT.collection_field,
                    match=models.MatchValue(value=collection_name),
                ),
            ]
            if collection_name is not None
            else []  # for mypy
        )
    )


async def delete_vectors(
    qdrant_collection_name: str, collection_name: str | None = None, document_id: uuid.UUID | None = None
) -> None:
    if client is None:
        raise ValueError("qdrant_client not set")
    await client.delete(
        collection_name=qdrant_collection_name,
        points_selector=models.FilterSelector(filter=get_filter(collection_name, document_id)),
    )
    return None


async def update_meta(
    qdrant_collection_name: str,
    meta: dict[str, Any],
    collection_name: str | None = None,
    document_id: uuid.UUID | None = None,
) -> None:
    if client is None:
        raise ValueError("qdrant_client not set")
    await client.set_payload(
        collection_name=qdrant_collection_name,
        key="metadata",
        payload=meta,
        points=models.FilterSelector(filter=get_filter(collection_name, document_id)),
    )
    return None


async def update_collection_name(qdrant_collection_name: str, collection_name: str, new_collection_name: str) -> None:
    await update_meta(
        qdrant_collection_name, {settings.QDRANT.collection_field: new_collection_name}, collection_name=collection_name
    )
    return None
