import base64
import json
import os
from collections.abc import Callable, Hashable
from datetime import datetime, timedelta
from enum import Enum
from typing import Any
from uuid import uuid4

from fastapi import (
    APIRouter,
    Depends,
    Header,
    HTTPException,
    Path,
    Request,
    Response,
    status,
)
from pydantic import BaseModel
from starlette.responses import FileResponse

from app.api.deps import CurrentUser, SessionDep
from app.core.config import settings

"""
Taken from : https://github.com/edihasaj/tuspy-fast-api/blob/aec7ba5ae548ea17c97cf203a047c6056ee8ee05/tusserver/tus.py
And modified to include tags and API prefix
"""


class FileMetadata(BaseModel):
    uid: str
    metadata: dict[Hashable, str]
    size: int
    offset: int = 0
    upload_part: int = 0
    created_at: str
    defer_length: bool
    upload_chunk_size: int = 0
    expires: str | None

    @classmethod
    def from_request(
        cls,
        uid: str,
        metadata: dict[Any, str],
        size: int,
        created_at: str,
        defer_length: bool,
        expires: str,
    ) -> "FileMetadata":
        return FileMetadata(
            uid=uid,
            metadata=metadata,
            size=size,
            created_at=created_at,
            defer_length=defer_length,
            expires=expires,
        )


def create_api_router(
    files_dir: str = "/tmp/files",
    max_size: int = 128849018880,
    on_upload_complete: Callable[[SessionDep, CurrentUser, str, dict[Any, Any], int], None] | None = None,
    days_to_keep: int = 5,
    prefix: str = "files",
    tags: list[str | Enum] | None = None,
) -> APIRouter:
    router = APIRouter(prefix=f"/{prefix}", tags=tags)

    tus_version = "1.0.0"
    tus_extension = "creation,creation-defer-length,creation-with-upload,expiration,termination"

    async def _get_request_chunk(request: Request, uuid: str = Path(...), post_request: bool = False) -> bool | None:
        meta = _read_metadata(uuid)
        if not meta or not _file_exists(uuid):
            return False

        # save the file as pdf since we only manage .pdf
        with open(f"{files_dir}/{uuid}.pdf", "ab") as f:
            async for chunk in request.stream():
                if post_request and chunk is None or len(chunk) == 0:
                    return None

                if _get_file_length(uuid) + len(chunk) > max_size:
                    raise HTTPException(
                        status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE,
                        detail=f"Document size must be less than {max_size} bytes.",
                    )

                f.write(chunk)
                meta.offset += len(chunk)
                meta.upload_chunk_size = len(chunk)
                meta.upload_part += 1
                _write_metadata(meta)

            f.close()

        return True

    @router.head("/{uuid}", status_code=status.HTTP_200_OK)
    def get_upload_metadata(_: CurrentUser, response: Response, uuid: str) -> Response:
        meta = _read_metadata(uuid)
        if meta is None or not _file_exists(uuid):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

        response.headers["Tus-Resumable"] = tus_version
        response.headers["Content-Length"] = str(meta.size)
        response.headers["Upload-Length"] = str(meta.size)
        response.headers["Upload-Offset"] = str(meta.offset)
        response.headers["Cache-Control"] = "no-store"
        encoded_filename = base64.b64encode(meta.metadata["filename"].encode("utf-8")).decode("utf-8")
        encoded_filetype = base64.b64encode(meta.metadata["filetype"].encode("utf-8")).decode("utf-8")
        response.headers["Upload-Metadata"] = f"filename {encoded_filename}, filetype {encoded_filetype}"
        response.status_code = status.HTTP_200_OK
        return response

    @router.patch("/{uuid}", status_code=status.HTTP_204_NO_CONTENT)
    def upload_chunk(
        response: Response,
        session: SessionDep,
        current_user: CurrentUser,
        uuid: str,
        content_type: str = Header(None),
        content_length: int = Header(None),
        upload_offset: int = Header(None),
        _: Any = Depends(_get_request_chunk),
    ) -> Response:
        response_headers = _get_and_save_the_file(
            response,
            session,
            current_user,
            uuid,
            content_type,
            content_length,
            upload_offset,
        )

        return response_headers

    @router.options("/", status_code=status.HTTP_204_NO_CONTENT)
    def options_create_upload(_: CurrentUser, response: Response) -> Response:
        response.headers["Tus-Extension"] = tus_extension
        response.headers["Tus-Resumable"] = tus_version
        response.headers["Tus-Version"] = tus_version
        response.headers["Tus-Max-Size"] = str(max_size)
        response.headers["Content-Length"] = str(0)
        response.status_code = status.HTTP_204_NO_CONTENT
        return response

    @router.post("/", status_code=status.HTTP_201_CREATED)
    async def create_upload(
        request: Request,
        response: Response,
        session: SessionDep,
        current_user: CurrentUser,
        upload_metadata: str = Header(None),
        upload_length: int = Header(None),
        upload_defer_length: int = Header(None),
    ) -> Response:
        if upload_defer_length is not None and upload_defer_length != 1:
            raise HTTPException(status_code=400, detail="Invalid Upload-Defer-Length")

        defer_length = upload_defer_length is not None

        # Create a new upload and store the file and metadata in the mapping
        metadata = {}
        if upload_metadata is not None and upload_metadata != "":
            # Decode the base64-encoded string
            for kv in upload_metadata.split(","):
                key, value = kv.rsplit(" ", 1)
                decoded_value = base64.b64decode(value.strip()).decode("utf-8")
                metadata[key.strip()] = decoded_value

        for key in ["filetype", "filename", "collection_id"]:
            if key not in metadata:
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f"{key} must be in the metadata",
                )

        if "filetype" not in metadata or metadata["filetype"] != "application/pdf":
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="We only accept pdf file for upload",
            )

        uuid = str(uuid4().hex)

        date_expiry = datetime.now() + timedelta(days=days_to_keep)
        saved_meta_data = FileMetadata.from_request(
            uuid,
            metadata,
            upload_length,
            str(datetime.now()),
            defer_length,
            str(date_expiry.isoformat()),
        )
        _write_metadata(saved_meta_data)

        _initialize_file(uuid)

        chunk: bool | None = await _get_request_chunk(request, uuid, True)
        if chunk:
            response = _get_and_save_the_file(
                response,
                session,
                current_user,
                uuid,
            )
            response.headers["Location"] = _build_location_url(request=request, uuid=uuid)
            return response

        response.headers["Location"] = _build_location_url(request=request, uuid=uuid)
        response.headers["Tus-Resumable"] = tus_version
        response.headers["Content-Length"] = str(0)
        response.status_code = status.HTTP_201_CREATED
        return response

    @router.options("/{uuid}", status_code=status.HTTP_204_NO_CONTENT)
    def options_upload_chunk(_: CurrentUser, response: Response, uuid: str) -> Response:
        meta = _read_metadata(uuid)
        if meta is None or not _file_exists(uuid):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

        response.headers["Tus-Extension"] = tus_extension
        response.headers["Tus-Resumable"] = tus_version
        response.headers["Tus-Version"] = tus_version
        response.headers["Content-Length"] = str(0)
        response.status_code = status.HTTP_204_NO_CONTENT
        return response

    @router.get("/{uuid}")
    def get_upload(_: CurrentUser, uuid: str) -> FileResponse:
        meta = _read_metadata(uuid)

        # Check if the upload ID is valid
        if not meta or uuid != meta.uid or not _file_exists(uuid):
            raise HTTPException(status_code=404, detail="Upload not found")

        # Return the file in the response
        return FileResponse(
            os.path.join(files_dir, uuid),
            media_type="application/octet-stream",
            filename=meta.metadata["filename"],
            headers={"Content-Length": str(meta.offset), "Tus-Resumable": tus_version},
        )

    @router.delete("/{uuid}", status_code=status.HTTP_204_NO_CONTENT)
    def delete_upload(_: CurrentUser, uuid: str, response: Response) -> Response:
        meta = _read_metadata(uuid)

        # Check if the upload ID is valid
        if not meta or uuid != meta.uid or not _file_exists(uuid):
            raise HTTPException(status_code=404, detail="Upload not found")

        # Delete the file and metadata for the upload from the mapping
        _delete_files(uuid)

        # Return a 204 No Content response
        response.headers["Tus-Resumable"] = tus_version
        response.status_code = status.HTTP_204_NO_CONTENT
        return response

    def _write_metadata(meta: FileMetadata) -> None:
        if not os.path.exists(files_dir):
            os.makedirs(files_dir)

        with open(os.path.join(files_dir, f"{meta.uid}.info"), "w") as f:
            f.write(json.dumps(meta, indent=4, default=lambda k: k.__dict__))

    def _initialize_file(uid: str) -> None:
        if not os.path.exists(files_dir):
            os.makedirs(files_dir)

        open(os.path.join(files_dir, f"{uid}"), "a").close()

    def _read_metadata(uid: str) -> FileMetadata | None:
        fpath = os.path.join(files_dir, f"{uid}.info")
        if os.path.exists(fpath):
            with open(fpath) as f:
                return FileMetadata(**json.load(f))

        return None

    def _get_file(uid: str) -> bytes | None:
        fpath = os.path.join(files_dir, uid)
        if os.path.exists(fpath):
            with open(fpath, "rb") as f:
                return f.read()

        return None

    def _file_exists(uid: str) -> bool:
        return os.path.exists(os.path.join(files_dir, uid))

    def _get_file_length(uid: str) -> int:
        return os.path.getsize(os.path.join(files_dir, uid))

    def _delete_files(uid: str) -> None:
        fpath = os.path.join(files_dir, f"{uid}.pdf")
        if os.path.exists(fpath):
            os.remove(fpath)

        meta_path = os.path.join(files_dir, f"{uid}.info")
        if os.path.exists(meta_path):
            os.remove(meta_path)

    def _get_and_save_the_file(
        response: Response,
        session: SessionDep,
        current_user: CurrentUser,
        uuid: str,
        content_type: str = Header(None),
        content_length: int = Header(None),
        upload_length: int = Header(None),
    ) -> Response:
        # Check if the Content-Type header is set to "application/offset+octet-stream"
        if content_type != "application/offset+octet-stream":
            raise HTTPException(status_code=415)

        meta = _read_metadata(uuid)
        # Check if the upload ID is valid
        if not meta or uuid != meta.uid:
            raise HTTPException(status_code=404)

        # Check if the Upload Offset with Content-Length header is correct
        if meta.offset != upload_length + content_length:
            raise HTTPException(status_code=409)

        if meta.defer_length:
            meta.size = upload_length

        if not meta.expires:
            date_expiry = datetime.now() + timedelta(days=days_to_keep)
            meta.expires = str(date_expiry.isoformat())
        _write_metadata(meta)

        if meta.size == meta.offset:
            response.headers["Tus-Resumable"] = tus_version
            response.headers["Upload-Offset"] = str(str(meta.offset) if meta.offset > 0 else str(content_length))
            response.headers["Upload-Expires"] = str(meta.expires)
            response.status_code = status.HTTP_204_NO_CONTENT
            if on_upload_complete:
                on_upload_complete(
                    session, current_user, os.path.join(files_dir, f"{uuid}.pdf"), meta.metadata, meta.size
                )
                # Remove the file and its metadata
                _delete_files(uuid)

            return response

        response.headers["Tus-Resumable"] = tus_version
        response.headers["Upload-Offset"] = str(meta.offset)
        response.headers["Upload-Expires"] = str(meta.expires)
        response.status_code = status.HTTP_204_NO_CONTENT
        return response

    def remove_expired_files() -> None:
        file_list = os.listdir(files_dir)

        file_list_to_remove = []

        for f in file_list:
            if len(f) == 32:
                file_list_to_remove.append(f)

        for f in file_list_to_remove:
            meta = _read_metadata(f)
            if meta and meta.expires and datetime.fromisoformat(meta.expires) < datetime.now():
                _delete_files(f)

    def _get_host_and_proto(request: Request) -> tuple[str, str | None]:
        proto = "http"
        host = request.headers.get("host")
        forward_proto = request.headers.get("X-Forwarded-Proto")
        if forward_proto is not None:
            proto = forward_proto
        if request.headers.get("X-Forwarded-Host") is not None:
            host = request.headers.get("X-Forwarded-Host")
        return proto, host

    def _build_location_url(request: Request, uuid: str) -> str:
        proto, host = _get_host_and_proto(request=request)
        return f"{proto}://{host}{settings.API_V1_STR}/{prefix}/{uuid}"

    # # Create a scheduler for deleting the files
    # scheduler = sched.scheduler(time.time, time.sleep)
    # run_time = time.mktime(time.strptime("01:00", "%H:%M"))
    # scheduler.enterabs(run_time, 1, remove_expired_files)
    # scheduler.run()

    return router
