from typing import Any

from fastapi.testclient import TestClient
from sqlmodel import Session, select

from app.core.config import settings
from app.models import ChatQuery
from app.tests.utils.collection import create_random_collection
from app.tests.utils.user import AuthenticatedUser, random_authenticated_user
from app.tests.utils.utils import random_lower_string


def create_random_query(client: TestClient, authent_user: AuthenticatedUser, db: Session) -> Any:
    collection = create_random_collection(db, authent_user.user)
    query = random_lower_string() + random_lower_string()
    data = {"query": query, "collection_id": str(collection.id)}
    response = client.post(
        f"{settings.API_V1_STR}/chat/query",
        headers=authent_user.token,
        json=data,
    )
    return response.json()


def test_store_query(client: TestClient, db: Session) -> None:
    authent_user = random_authenticated_user(db)
    query = create_random_query(client, authent_user, db)
    query_json = {"query": query["query"], "collection_ids": [query["collection_id"]]}
    response = client.post(
        f"{settings.API_V1_STR}/chat/query/{query['id']}/message",
        headers=authent_user.token,
        json=query_json,
    )
    assert response.status_code == 200

    chat_query = select(ChatQuery).where(ChatQuery.user_id == authent_user.user.id)
    chat_query_db = db.exec(chat_query).first()
    assert chat_query_db
    assert str(chat_query_db.collection_id) == query["collection_id"]
    assert chat_query_db.query == query["query"]
    assert chat_query_db.title == query["title"]
    # Long title are croped
    assert chat_query_db.title[-3:] == "..."
    assert chat_query_db.history == []


def test_chat_queries(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    authent_user_1 = random_authenticated_user(db)
    authent_user_2 = random_authenticated_user(db)
    query_1 = create_random_query(client, authent_user_1, db)
    query_2 = create_random_query(client, authent_user_2, db)

    # Only superadmin can access the route
    response = client.get(
        f"{settings.API_V1_STR}/chat/queries",
        headers=authent_user_1.token,
    )
    assert response.status_code == 403

    response = client.get(
        f"{settings.API_V1_STR}/chat/queries",
        headers=superuser_token_headers,
    )
    assert response.status_code == 200
    content = response.json()
    assert len(content["items"]) == 2
    items = content["items"]
    assert items[1]["query"] == query_1["query"]
    assert items[1]["user_id"] == str(authent_user_1.user.id)
    assert "history" not in items[0]
    assert items[0]["query"] == query_2["query"]
    assert items[0]["user_id"] == str(authent_user_2.user.id)


def test_chat_history() -> None:
    pass
