import os
import uuid

import pytest
from fastapi import status
from fastapi.testclient import TestClient
from sqlmodel import Session

from app.core.config import settings
from app.main import PROJECT_PATH
from app.models import (
    CollectionType,
    UserRole,
)
from app.tests.utils.collection import create_random_collection
from app.tests.utils.permission import UserPerm, user_generator
from app.tests.utils.user import random_authenticated_user
from app.tests.utils.utils import random_lower_string


def test_create_collection(client: TestClient, superuser_token_headers: dict[str, str]) -> None:
    data = {"title": "Foo", "description": "Fighters"}
    response = client.post(
        f"{settings.API_V1_STR}/collections/",
        headers=superuser_token_headers,
        json=data,
    )
    assert response.status_code == status.HTTP_200_OK
    content = response.json()
    assert content["title"] == data["title"]
    assert content["description"] == data["description"]
    assert "id" in content
    assert "user_id" in content


def test_create_duplicate_collection(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    collection = create_random_collection(db)
    data = {"title": collection.title}
    response = client.post(
        f"{settings.API_V1_STR}/collections/",
        headers=superuser_token_headers,
        json=data,
    )
    assert response.status_code == 400
    assert response.json() == {"detail": "A collection with the given name already exists"}


def test_read_collection(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    collection = create_random_collection(db)
    response = client.get(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=superuser_token_headers,
    )
    assert response.status_code == status.HTTP_200_OK
    content = response.json()
    assert content["title"] == collection.title
    assert content["description"] == collection.description
    assert content["id"] == str(collection.id)
    assert content["user_id"] == str(collection.user_id)


def test_read_collection_not_found(client: TestClient, superuser_token_headers: dict[str, str]) -> None:
    response = client.get(
        f"{settings.API_V1_STR}/collections/{uuid.uuid4()}",
        headers=superuser_token_headers,
    )
    assert response.status_code == 404
    content = response.json()
    assert content["detail"] == "Collection not found"


def test_read_collection_not_enough_permissions(
    client: TestClient, normal_user_token_headers: dict[str, str], db: Session
) -> None:
    collection = create_random_collection(db)
    response = client.get(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=normal_user_token_headers,
    )
    assert response.status_code // 100 == 4
    content = response.json()
    assert content["detail"] == "The user doesn't have enough privileges"


def test_read_collections(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    create_random_collection(db)
    create_random_collection(db)
    response = client.get(
        f"{settings.API_V1_STR}/collections/",
        headers=superuser_token_headers,
    )
    assert response.status_code == status.HTTP_200_OK
    content = response.json()
    assert len(content["items"]) >= 2


def test_update_collection(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    collection = create_random_collection(db)
    # data = {"title": "Updated title", "description": "Updated description"}
    data = {"description": "Updated description"}
    response = client.put(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=superuser_token_headers,
        json=data,
    )
    assert response.status_code == status.HTTP_200_OK
    content = response.json()
    # assert content["title"] == data["title"]
    assert content["description"] == data["description"]
    assert content["id"] == str(collection.id)
    assert content["user_id"] == str(collection.user_id)


def test_update_collection_title(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    collection = create_random_collection(db)
    data = {"title": "Updated title", "description": "Updated description"}
    response = client.put(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=superuser_token_headers,
        json=data,
    )
    assert response.status_code == 501


def test_update_collection_not_found(client: TestClient, superuser_token_headers: dict[str, str]) -> None:
    # data = {"title": "Updated title", "description": "Updated description"}
    data = {"description": "Updated description"}
    response = client.put(
        f"{settings.API_V1_STR}/collections/{uuid.uuid4()}",
        headers=superuser_token_headers,
        json=data,
    )
    assert response.status_code == 404
    content = response.json()
    assert content["detail"] == "Collection not found"


def test_update_collection_not_enough_permissions(
    client: TestClient, normal_user_token_headers: dict[str, str], db: Session
) -> None:
    collection = create_random_collection(db)
    # data = {"title": "Updated title", "description": "Updated description"}
    data = {"description": "Updated description"}
    response = client.put(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=normal_user_token_headers,
        json=data,
    )
    assert response.status_code // 100 == 4
    content = response.json()
    assert content["detail"] == "The user doesn't have enough privileges"


def test_delete_collection(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    collection = create_random_collection(db)
    response = client.delete(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=superuser_token_headers,
    )
    assert response.status_code == status.HTTP_200_OK
    content = response.json()
    assert content["message"] == "Collection deleted successfully"


def test_delete_collection_not_found(client: TestClient, superuser_token_headers: dict[str, str]) -> None:
    response = client.delete(
        f"{settings.API_V1_STR}/collections/{uuid.uuid4()}",
        headers=superuser_token_headers,
    )
    assert response.status_code == 404
    content = response.json()
    assert content["detail"] == "Collection not found"


def test_delete_collection_not_enough_permissions(
    client: TestClient, normal_user_token_headers: dict[str, str], db: Session
) -> None:
    collection = create_random_collection(db)
    response = client.delete(
        f"{settings.API_V1_STR}/collections/{collection.id}",
        headers=normal_user_token_headers,
    )
    assert response.status_code // 100 == 4
    content = response.json()
    assert content["detail"] == "The user doesn't have enough privileges"


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_200_OK,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_200_OK,
                UserPerm.other_group_admin: status.HTTP_200_OK,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_200_OK,
            },
        ),
    ],
)
def test_create_permission_on_collection(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    group_admin = random_authenticated_user(test_db, role=creator_role, group_id=1)

    for token, expected_status in user_generator(test_db, group_admin, exp_usr_status):
        data = {"title": random_lower_string(), "type": collection_type.value}
        response = client.post(
            f"{settings.API_V1_STR}/collections/",
            headers=token,
            json=data,
        )
        assert response.status_code == expected_status


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_200_OK,
                UserPerm.other_group_admin: status.HTTP_200_OK,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_200_OK,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_read_permission_on_collection(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        response = client.get(f"{settings.API_V1_STR}/collections/{collection.id}", headers=token)
        assert response.status_code == expected_status


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_update_permission_on_collection(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        response = client.put(
            f"{settings.API_V1_STR}/collections/{collection.id}",
            headers=token,
            json={"description": "Bar"},
        )
        assert response.status_code == expected_status


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_admin: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_delete_permission_on_collection(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        response = client.delete(f"{settings.API_V1_STR}/collections/{collection.id}", headers=token)
        assert response.status_code == expected_status
        if response.status_code == status.HTTP_200_OK:
            collection = create_random_collection(test_db, collection_creator.user, collection_type)


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_create_document_permission_on_collection(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        with open(os.path.join(PROJECT_PATH, "fixtures", "dummy.pdf"), "rb") as f:
            response = client.post(
                f"{settings.API_V1_STR}/collections/{collection.id}/document", headers=token, files={"files": f}
            )
            assert response.status_code == expected_status
