import pytest
from fastapi import status
from fastapi.testclient import TestClient
from sqlmodel import Session

from app.core.config import CollectionType, settings
from app.models import UserRole
from app.tests.utils.collection import create_random_collection
from app.tests.utils.documents import create_dummy_document
from app.tests.utils.permission import UserPerm, user_generator
from app.tests.utils.user import random_authenticated_user
from app.tests.utils.utils import random_lower_string


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_403_FORBIDDEN,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_200_OK,
                UserPerm.other_group_admin: status.HTTP_200_OK,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_403_FORBIDDEN,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_admin: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_403_FORBIDDEN,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_200_OK,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_read_document_permission(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)
    document_id = create_dummy_document(client, collection, collection_creator.token)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        # We have to send token through query param for this one
        # We use split because header token contains "Bearer "
        token_param = "" if "Authorization" not in token else token["Authorization"].split(" ")[1]
        response = client.get(f"{settings.API_V1_STR}/documents/{document_id}/file?token={token_param}")
        assert response.status_code == expected_status


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_admin: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_update_document_permission(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)
    document_id = create_dummy_document(client, collection, collection_creator.token)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        # We have to send token through query param for this one
        # We use split because header token contains "Bearer "
        response = client.put(
            f"{settings.API_V1_STR}/documents/{document_id}",
            headers=token,
            json={"title": random_lower_string(), "meta": {}, "comments": {}},
        )
        assert response.status_code == expected_status


@pytest.mark.parametrize(
    "collection_type,creator_role,exp_usr_status",
    [
        (
            CollectionType.public,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.personal,
            UserRole.user,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_admin: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_403_FORBIDDEN,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
        (
            CollectionType.restricted,
            UserRole.groupadmin,
            {
                UserPerm.none: status.HTTP_401_UNAUTHORIZED,
                UserPerm.superadmin: status.HTTP_200_OK,
                UserPerm.creator: status.HTTP_200_OK,
                UserPerm.group_contributor: status.HTTP_200_OK,
                UserPerm.group_member: status.HTTP_403_FORBIDDEN,
                UserPerm.other_group_admin: status.HTTP_403_FORBIDDEN,
            },
        ),
    ],
)
def test_delete_document_permission(
    client: TestClient,
    test_db: Session,
    collection_type: CollectionType,
    creator_role: UserRole,
    exp_usr_status: dict[UserPerm, int],
) -> None:
    collection_creator = random_authenticated_user(test_db, role=creator_role, group_id=1)
    collection = create_random_collection(test_db, collection_creator.user, collection_type)
    document_id = create_dummy_document(client, collection, collection_creator.token)

    for token, expected_status in user_generator(test_db, collection_creator, exp_usr_status):
        # We have to send token through query param for this one
        # We use split because header token contains "Bearer "
        response = client.delete(
            f"{settings.API_V1_STR}/documents/{document_id}",
            headers=token,
        )
        assert response.status_code == expected_status
        if expected_status == 200:
            # We need to recreate the document
            document_id = create_dummy_document(client, collection, collection_creator.token)
