import uuid

import pytest
from fastapi.testclient import TestClient
from sqlmodel import Session

from app.core.config import settings
from app.models import UserRole
from app.tests.utils.user import authentication_token_from_email, create_random_user


def test_get_access_token(client: TestClient) -> None:
    login_data = {
        "username": settings.FIRST_SUPERUSER,
        "password": settings.FIRST_SUPERUSER_PASSWORD,
    }
    r = client.post(f"{settings.API_V1_STR}/login/access-token", data=login_data)
    tokens = r.json()
    assert r.status_code == 200
    assert "access_token" in tokens
    assert tokens["access_token"]


def test_get_access_token_incorrect_password(client: TestClient) -> None:
    login_data = {
        "username": settings.FIRST_SUPERUSER,
        "password": "incorrect",
    }
    r = client.post(f"{settings.API_V1_STR}/login/access-token", data=login_data)
    assert r.status_code == 400


def test_use_access_token(client: TestClient, superuser_token_headers: dict[str, str]) -> None:
    r = client.post(
        f"{settings.API_V1_STR}/login/test-token",
        headers=superuser_token_headers,
    )
    result = r.json()
    assert r.status_code == 200
    assert "email" in result


def test_log_as(client: TestClient, superuser_token_headers: dict[str, str], db: Session) -> None:
    user = create_random_user(db)
    r = client.post(
        f"{settings.API_V1_STR}/login/log-as/{user.id}",
        headers=superuser_token_headers,
    )
    t = r.json()
    headers = {"Authorization": f"{t['token_type']} {t['access_token']}"}
    r = client.post(
        f"{settings.API_V1_STR}/login/test-token",
        headers=headers,
    )
    u = r.json()
    assert u["email"] == user.email

    wrong_uuid = uuid.uuid4()
    r = client.post(
        f"{settings.API_V1_STR}/login/log-as/{wrong_uuid}",
        headers=superuser_token_headers,
    )
    assert r.status_code == 404


@pytest.mark.parametrize(
    "user_role,expected_status_code",
    [
        (UserRole.groupadmin, 403),
        (UserRole.user, 403),
        (UserRole.contributor, 403),
        (UserRole.superadmin, 200),
    ],
)
def test_log_as_only_superuser(client: TestClient, db: Session, user_role: UserRole, expected_status_code: int) -> None:
    random_user = create_random_user(db)
    user = create_random_user(db, user_role)
    auth_token = authentication_token_from_email(client=client, email=user.email, db=db)
    r = client.post(
        f"{settings.API_V1_STR}/login/log-as/{random_user.id}",
        headers=auth_token,
    )
    assert r.status_code == expected_status_code
