import uuid
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any
from urllib.parse import urlparse, urlunparse

import pytest
from fastapi.testclient import TestClient
from pydantic import AnyUrl
from sqlalchemy import Engine, create_engine
from sqlalchemy_utils import create_database, drop_database
from sqlmodel import Session, SQLModel

from app.api.deps import get_db
from app.core.config import settings
from app.main import app
from app.tests.test_data import create_test_data
from app.tests.utils.user import authentication_token_from_email
from app.tests.utils.utils import get_superuser_token_headers


def modify_url_path(url: AnyUrl, new_path: str) -> AnyUrl:
    parsed_url = urlparse(str(url))
    modified_url = parsed_url._replace(path=new_path)
    return AnyUrl(urlunparse(modified_url))


@contextmanager
def tmp_database(db_url: AnyUrl, suffix: str = "", **kwargs: Any) -> Generator[str, None, None]:
    tmp_db_name = ".".join([uuid.uuid4().hex, "sofia_collection", suffix])
    tmp_db_url = str(modify_url_path(db_url, tmp_db_name))
    create_database(tmp_db_url, **kwargs)

    try:
        yield tmp_db_url
    finally:
        drop_database(tmp_db_url)


@pytest.fixture(scope="session")
def postgres() -> Generator[str, None, None]:
    """
    Creates empty temporary database.
    """
    with tmp_database(settings.SQLALCHEMY_PG_URI, "pytest") as tmp_url:
        yield tmp_url


@pytest.fixture(scope="function")
def migration_postgres() -> Generator[str, None, None]:
    """
    Creates empty temporary database.
    Need to be only use for migration since other tests depends on recreate_db
    which depends on "postgres" fixture and this mess with migration testing
    """
    with tmp_database(settings.SQLALCHEMY_PG_URI, "pytest") as tmp_url:
        yield tmp_url


@pytest.fixture(scope="session")
def postgres_engine(postgres: str) -> Generator[Engine, None, None]:
    """
    SQLAlchemy engine, bound to temporary database.
    """
    # Set echo to true for test debugging
    engine = create_engine(postgres, echo=False)
    try:
        yield engine
    finally:
        engine.dispose()


@pytest.fixture(scope="session", autouse=True)
def recreate_db(postgres_engine: Engine) -> None:
    with Session(postgres_engine) as session:
        SQLModel.metadata.create_all(postgres_engine)
        create_test_data(session)  # TODO: create groups and users


# I didn't find a more elegant way to share db session with TestClient and with tests themself
# (some tests need to check that db contains some objects)
# They need the same data and because everything is in a transaction (which is rollback after each test)
# We can't create another session for the tests
current_session: Session | None = None


@pytest.fixture(scope="function")
def test_db(postgres_engine: Engine) -> Generator[Session, None, None]:
    connection = postgres_engine.connect()
    connection.begin()
    session = Session(bind=connection)

    global current_session
    current_session = session
    yield session

    session.rollback()
    connection.close()


@pytest.fixture(scope="function")
def db() -> Session | None:
    global current_session
    return current_session


@pytest.fixture(scope="function")
def client(test_db: Session) -> Generator[TestClient, None, None]:
    def override_get_db() -> Generator[Session, None, None]:
        yield test_db

    app.dependency_overrides[get_db] = override_get_db
    with TestClient(app) as c:
        yield c
    app.dependency_overrides.clear()


@pytest.fixture(scope="function")
def superuser_token_headers(client: TestClient) -> dict[str, str]:
    return get_superuser_token_headers(client)


@pytest.fixture(scope="function")
def normal_user_token_headers(client: TestClient, db: Session) -> dict[str, str]:
    return authentication_token_from_email(client=client, email=settings.EMAIL_TEST_USER, db=db)


# @pytest.fixture(scope="function")
# def group_session(session: Session) -> Generator[Session, None, None]:
#     for entry in GROUP_TABLE:
#         session.add(Group(name=random_lower_string(), **entry))
#     session.commit()
#     yield session
#     session.rollback()


# @pytest.fixture(scope="function")
# def get_user_id(user_idx: int) -> uuid.UUID:
#     return cast(uuid.UUID, USER_TABLE[user_idx]["id"])
