from fastapi.encoders import jsonable_encoder
from sqlmodel import Session

from app import crud
from app.core.security import verify_password
from app.models import User, UserCreate, UserRole, UserUpdate
from app.tests.utils.utils import create_random_user, random_email, random_lower_string


def test_create_user(test_db: Session) -> None:
    email = random_email()
    user = create_random_user(test_db, email=email)
    assert user.email == email
    assert hasattr(user, "hashed_password")


def test_authenticate_user(test_db: Session) -> None:
    email = random_email()
    password = random_lower_string()
    user_in = UserCreate(email=email, password=password, group_id=1)
    user = crud.create_user(session=test_db, user_create=user_in)
    authenticated_user = crud.authenticate(session=test_db, email=email, password=password)
    assert authenticated_user
    assert user.email == authenticated_user.email


def test_not_authenticate_user(test_db: Session) -> None:
    email = random_email()
    password = random_lower_string()
    user = crud.authenticate(session=test_db, email=email, password=password)
    assert user is None


def test_check_if_user_is_active(test_db: Session) -> None:
    user = create_random_user(test_db)
    assert user.is_active is True


def test_check_if_user_is_active_inactive(test_db: Session) -> None:
    user = create_random_user(test_db, disabled=True)
    assert user.is_active


def test_check_if_user_is_superuser(test_db: Session) -> None:
    user = create_random_user(test_db, role=UserRole.superadmin)
    assert user.role == UserRole.superadmin


def test_check_if_user_is_superuser_normal_user(test_db: Session) -> None:
    username = random_email()
    password = random_lower_string()
    user_in = UserCreate(email=username, password=password, group_id=1)
    user = crud.create_user(session=test_db, user_create=user_in)
    assert user.role != UserRole.superadmin


def test_get_user(test_db: Session) -> None:
    user = create_random_user(test_db)
    user_2 = test_db.get(User, user.id)
    assert user_2
    assert user.email == user_2.email
    assert jsonable_encoder(user) == jsonable_encoder(user_2)


def test_update_user(test_db: Session) -> None:
    user = create_random_user(test_db)
    new_password = random_lower_string()
    user_in_update = UserUpdate(password=new_password)
    if user.id is not None:
        crud.update_user(session=test_db, db_user=user, user_in=user_in_update)
    user_2 = test_db.get(User, user.id)
    assert user_2
    assert user.email == user_2.email
    assert user_2.hashed_password
    assert verify_password(new_password, user_2.hashed_password)
