import importlib
import os
import uuid
from collections import defaultdict, namedtuple
from typing import Any

import pytest
from alembic.config import Config
from sqlmodel import Session, text

from app.main import PROJECT_PATH


def create_group(session: Session) -> int:
    group_id = 1
    query = text(f"""insert into public.group (id, name, domain) values ({group_id}, 'test', 'test.com')""")
    session.execute(query)
    return group_id


def create_user(session: Session, group_id: int) -> uuid.UUID:
    user_id = uuid.uuid4()
    query = text(f"""insert into public.user (id, email, is_active, role, sso, group_id)
    values ('{user_id}', 'test@test.com', true, 'user', true, {group_id})""")
    session.execute(query)
    return user_id


def create_collection(session: Session, group_id: int, user_id: uuid.UUID) -> uuid.UUID:
    collection_id = uuid.uuid4()
    query = text(f"""insert into public.collection (id, title, type, user_id, group_id, created_at)
    values ('{collection_id}', 'test', 'personal', '{user_id}', {group_id}, current_timestamp);""")
    session.execute(query)
    return collection_id


def load_migration_as_module(file: str) -> Any:
    """
    Allows to import alembic migration as a module.
    """
    return importlib.machinery.SourceFileLoader(
        file, os.path.join(PROJECT_PATH, "app", "alembic", "versions", file)
    ).load_module()


# Represents test for 'data' migration.
# Contains revision to be tested, it's previous revision, and callbacks that
# could be used to perform validation.
MigrationValidationParamsGroup = namedtuple(
    "MigrationValidationParamsGroup", ["rev_base", "rev_head", "on_init", "on_upgrade", "on_downgrade"]
)


def make_validation_params_groups(*migrations: Any) -> list[MigrationValidationParamsGroup]:
    """
    Creates objects that describe test for data migrations.
    See examples in tests/data_migrations/migration_*.py.
    """
    data = []
    for migration in migrations:
        # Ensure migration has all required params
        for required_param in ["rev_base", "rev_head"]:
            if not hasattr(migration, required_param):
                raise RuntimeError(f"{required_param} not specified for {migration.__name__}")

        # Set up callbacks
        callbacks: dict[str, Any] = defaultdict(lambda: lambda *args, **kwargs: None)
        for callback in ["on_init", "on_upgrade", "on_downgrade"]:
            if hasattr(migration, callback):
                callbacks[callback] = getattr(migration, callback)

        data.append(
            MigrationValidationParamsGroup(
                rev_base=migration.rev_base,
                rev_head=migration.rev_head,
                on_init=callbacks["on_init"],
                on_upgrade=callbacks["on_upgrade"],
                on_downgrade=callbacks["on_downgrade"],
            )
        )

    return data


def make_alembic_config(postgres_url: str | None = None) -> Config:
    config = Config(file_=os.path.join(PROJECT_PATH, "alembic.ini"), ini_section="alembic")

    # Replace path to alembic folder to absolute
    alembic_location = config.get_main_option("script_location")
    if alembic_location and not os.path.isabs(alembic_location):
        config.set_main_option("script_location", os.path.join(PROJECT_PATH, alembic_location))

    if postgres_url:
        config.set_main_option("sqlalchemy.url", postgres_url)
    return config


@pytest.fixture()
def alembic_config(migration_postgres: str) -> Config:
    """
    Alembic configuration object, bound to temporary database.
    """
    return make_alembic_config(migration_postgres)
