"""
Data migration test example.

MUST-have properties:
  - rev_base
    Previous migration identifier
  - rev_head
    Current (being tested) migration identifier

Optional properties:
  - on_init(engine)
    Is called before applying 'rev_head' migration. Can be used to add some
    data before migration is applied.
  - on_upgrade(engine)
    Is called after applying 'rev_head' migration. Can be used to check data
    was migrated successfully by upgrade() migration method.
  - on_downgrade(engine)
    Is called after migration is rolled back to 'rev_base'. Can be used to
    check data was rolled back to initial state.
"""

import uuid

from sqlalchemy.engine import Engine
from sqlmodel import Session, text

from app.tests.migrations.conftest import create_collection, create_group, create_user, load_migration_as_module

# Load migration as module
migration = load_migration_as_module("11d362313208_document_comments_column_is_now_not_.py")
rev_base = migration.down_revision
rev_head: str = migration.revision

document_id = uuid.uuid4()


def on_init(engine: Engine) -> None:
    """
    Create rows in users table before migration is applied
    """
    global document_id

    with Session(engine) as session:
        group_id = create_group(session)
        user_id = create_user(session, group_id)
        collection_id = create_collection(session, group_id, user_id)
        query = text(f"""insert into public.document(id, filename, content_length, collection_id, created_at, updated_at)
        values('{document_id}', 'test.pdf', 1000, '{collection_id}', current_timestamp, current_timestamp);""")
        session.execute(query)
        session.commit()


def on_upgrade(engine: Engine) -> None:
    """
    Ensure that data was successfully migrated
    """
    global document_id

    with Session(engine) as session:
        query = text(f"select * from public.document where id='{document_id}'")
        res = session.execute(query).first()
        assert res and res.comments is not None


def on_downgrade(engine: Engine) -> None:
    """
    Ensure that data changes were rolled back
    """
    # No test on this migration
