import logging
import uuid

from sqlmodel import Session, select

from app import crud
from app.core.config import settings
from app.core.db import engine
from app.models import Collection, CollectionCreate, Group, GroupBase, User, UserCreate

groups = {
    "default": {"id": 1, "name": settings.DEFAULT_GROUP},
    "ademe": {"id": 2, "name": "ADEME", "domain": "ademe.fr"},
    "cerema": {"id": 3, "name": "CEREMA", "domain": "cerema.fr"},
    "mte": {"id": 4, "name": "MTE", "domain": "developpement-durable.gouv.fr"},
}

users = {
    "superadmin": {
        "id": uuid.uuid4(),
        "group_id": 1,
        "role": "superadmin",
        "email": settings.FIRST_SUPERUSER,
        "password": settings.FIRST_SUPERUSER_PASSWORD,
    },
    "ademe_admin": {
        "id": uuid.uuid4(),
        "group_id": 2,
        "role": "groupadmin",
        "email": "admin@ademe.fr",
        "password": "testthis",
    },
    "ademe_contributor": {
        "id": uuid.uuid4(),
        "group_id": 2,
        "role": "contributor",
        "email": "contributor@ademe.fr",
        "password": "testthis",
    },
    "ademe_user": {"id": uuid.uuid4(), "group_id": 2, "role": "user", "email": "user@ademe.fr", "password": "testthis"},
    "cerema_admin": {
        "id": uuid.uuid4(),
        "group_id": 3,
        "role": "groupadmin",
        "email": "admin@cerema.fr",
        "password": "testthis",
    },
}

collections = {
    "default_public": {"title": "Publique", "type": "public", "group_id": groups["default"]["id"], "user_id": None},
    "ademe_public": {"title": "ADEME publique", "type": "public", "group_id": groups["ademe"]["id"], "user_id": None},
    "ademe_restricted": {
        "title": "ADEME restreinte",
        "type": "restricted",
        "group_id": groups["ademe"]["id"],
        "user_id": None,
    },
    "ademe_personnal": {
        "title": "Personal 1",
        "type": "personal",
        "group_id": groups["ademe"]["id"],
        "user_id": users["ademe_user"]["id"],
    },
    "cerema_personnal": {
        "title": "Personal 2",
        "type": "personal",
        "group_id": groups["cerema"]["id"],
        "user_id": users["cerema_admin"]["id"],
    },
}

# DOCUMENT_TABLE = [
#     {"id": uuid.uuid4(), "collection_id": 1},
#     {"id": uuid.uuid4(), "collection_id": 2},
# ]


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def create_test_data(session: Session) -> None:
    users_replace = {}
    for item in groups.values():
        if not session.get(Group, item["id"]):
            crud.create_generic_item(session=session, model=Group, item=GroupBase(**item))
    for item in users.values():
        if not session.exec(select(User).where(User.email == item["email"])).first():
            user = crud.create_user(session=session, user_create=UserCreate(**item))
            users_replace[item["id"]] = user.id

    for item in collections.values():
        # if not session.exec(select(Collection).where(Collection.type == item["t"]), item["id"]).first():
        crud.create_generic_item(
            session=session,
            model=Collection,
            item=CollectionCreate(**item),
            user_id=users_replace.get(item["user_id"]),
            group_id=item["group_id"],
        )


def main() -> None:
    logger.info("Creating test data")
    with Session(engine) as session:
        create_test_data(session)
    logger.info("Test data created")


if __name__ == "__main__":
    main()
