from datetime import timedelta

from fastapi.testclient import TestClient
from pydantic import BaseModel
from sqlmodel import Session

from app import crud
from app.core import security
from app.core.config import settings
from app.models import User, UserCreate, UserRole, UserUpdate
from app.tests.utils.utils import random_email, random_lower_string


def user_authentication_headers(*, client: TestClient, email: str, password: str) -> dict[str, str]:
    data = {"username": email, "password": password}

    r = client.post(f"{settings.API_V1_STR}/login/access-token", data=data)
    response = r.json()
    auth_token = response["access_token"]
    headers = {"Authorization": f"Bearer {auth_token}"}
    return headers


def create_random_user(db: Session, role: UserRole = UserRole.user, group_id: int = 1) -> User:
    email = random_email()
    password = random_lower_string()
    user_in = UserCreate(email=email, password=password, role=role, group_id=group_id)
    user = crud.create_user(session=db, user_create=user_in)
    return user


class AuthenticatedUser(BaseModel):
    user: User
    token: dict[str, str]


def random_authenticated_user(db: Session, role: UserRole = UserRole.user, group_id: int = 1) -> AuthenticatedUser:
    user = create_random_user(db, role=role, group_id=group_id)
    access_token_expires = timedelta(minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES)
    auth_token = security.create_access_token(user.id, expires_delta=access_token_expires)
    return AuthenticatedUser(user=user, token={"Authorization": f"Bearer {auth_token}"})


def authentication_token_from_email(
    *, client: TestClient, email: str, db: Session, role: UserRole = UserRole.user, group_id: int = 1
) -> dict[str, str]:
    """
    Return a valid token for the user with given email.

    If the user doesn't exist it is created first.
    """
    password = random_lower_string()
    user = crud.get_user_by_email(session=db, email=email)
    if not user:
        user_in_create = UserCreate(email=email, password=password, group_id=group_id, role=role)
        user = crud.create_user(session=db, user_create=user_in_create)
    else:
        user_in_update = UserUpdate(password=password)
        if not user.id:
            raise Exception("User id not set")
        user = crud.update_user(session=db, db_user=user, user_in=user_in_update)

    return user_authentication_headers(client=client, email=email, password=password)
